/******************************************************************************
*
*       CacheMan - Windows NT Cache Managing Tuner
*		
*		Copyright (c) 1997  Bryce Cogswell and Mark Russinovich
*
*    	PROGRAM: CacheMan.c
*
******************************************************************************/
#include <windows.h>
#include <stdio.h>
#include <process.h>
#include <commctrl.h>
#include <winsock.h>
#include <winioctl.h>
#include "resource.h"
#include "../dd/ioctlcmd.h"

//
// externs
//
BOOL LoadDeviceDriver( const char * Name, const char * Path, HANDLE * lphDevice );
BOOL UnloadDeviceDriver( const char * Name );

//
// instance
//
HINSTANCE			hInst;

//
// main window
//
HWND				hMainDlg;

//
// Variables/definitions for the driver that performs 
//
#define				SYS_FILE		TEXT("CACHEMAN.SYS")
#define				SYS_NAME		TEXT("CACHEMAN")
static HANDLE		sys_handle		= INVALID_HANDLE_VALUE;

//
// General buffer for storing temporary strings
//
static TCHAR		msgbuf[ 4096 ];

//
// forward define
//
void Abort( HWND hWnd, char * Msg );


#define	TESTING	0


//
// Tool tip help
//
static struct tip {
	int		ID;
	char *	Text;
} Tips[] = {
	{	IDC_GET,
		"Query system for current parameters and update dialog"	},
	{	IDC_SET,
		"Install specified parameters into system"	},

	{	IDC_CCFIRSTDELAY,
		"Delay before writing back a page after first access"	},
	{	IDC_CCIDLEDELAY,
		"Idle period to wait before writing back dirty pages"	},
	{	IDC_CCCOLLISIONDELAY, 
		"Delay by this amount if writeback not possible"		},
	{	IDC_CCTARGETCLEANDELAY,
		"Target clean delay"									},

	{	IDC_CCDIRTYPAGETHRESHOLD, 
		"Maximum number of dirty pages to allow in cache"		},
	{	IDC_CCDIRTYPAGETARGET,
		"Desired number of dirty pages in cache"				},
	{	IDC_CCAVAILABLEMAXDIRTYWRITE,
		"Maximum number of dirty pages to write back at a time"	},
	{	IDC_CCAVAILABLEPAGESTHRESHOLD,
		"Require at least this many pages available"	},

	{	IDC_MMSYSTEMCACHEWSMINIMUM,
		"Minimum size of the working set of the cache"	},
	{	IDC_MMSYSTEMCACHEWSMAXIMUM,
		"Maximum size of the working set of the cache"	},

	{	IDC_MMDOPERIODICAGRESSIVETRIMMING,
		"Agressively reduce working sets of all tasks"	},
	{	IDC_MMPERIODICAGRESSIVETRIMMINFREE,
		"Trim working set of cache if available pages greater than this amount"	},
	{	IDC_MMPERIODICAGRESSIVETRIMMAXFREE,
		"Trim working set of cache if available pages less than this amount"	},
	{	IDC_MMPERIODICAGRESSIVECACHEWSMIN,
		"Trim cache to this amount periodically"	},

	{	IDC_MMWORKINGSETREDUCTIONMAXCACHEWS,
		"Trim cache by this amount if larger than max size"	},
	{	IDC_MMWORKINGSETREDUCTIONMINCACHEWS,
		"Trim cache by this amount if smaller than max size"	},
	{	IDC_MMWORKINGSETVOLREDUCTIONMAXCACHEWS,
		"Trim cache by this amount if larger than max size during volume reduction"	},
};



//----------------------------------------------------------------------
//  
// DriverConnect
//
//----------------------------------------------------------------------
BOOL DriverConnect( void )
{
#if !TESTING
	TCHAR		Path[256], err[256];

	// open the handle to the device
	GetCurrentDirectory( sizeof Path, Path );
	wsprintf( Path+lstrlen(Path), TEXT("\\%s"), SYS_FILE );
	if ( ! LoadDeviceDriver( SYS_NAME, Path, &sys_handle ) )  {
		if( GetLastError() == 2 )
			sprintf( err, "File Not Found" );
		else
			sprintf( err, "Error %d", GetLastError() );
		wsprintf( msgbuf, TEXT("Opening %s (%s):%s"), SYS_NAME, Path,
			err);

		Abort( NULL, msgbuf );
		return FALSE;
	}	
#endif
	return TRUE;
}


//----------------------------------------------------------------------
//  
// DriverDisconnect
//
// Close the driver.
//
//----------------------------------------------------------------------
void DriverDisconnect()
{
#if !TESTING
	if( sys_handle != INVALID_HANDLE_VALUE ) {
		CloseHandle( sys_handle );			
		if ( ! UnloadDeviceDriver( SYS_NAME ) )  {
			wsprintf( msgbuf, TEXT("Error unloading \"%s\""), SYS_NAME );
			MessageBox( NULL, msgbuf, TEXT("NT Tune"), MB_OK );
		}	
	}
#endif
}

//----------------------------------------------------------------------
//  
// Abort
//
// If there is a fatal problem, pop up a message box and die. Assumes
// that it is called from the dialog box.
//
//----------------------------------------------------------------------
void Abort( HWND hWnd, char * Msg )
{
	MessageBox( hWnd, Msg, "Cache Manager", MB_ICONERROR|MB_OK );
	if ( hWnd ) 
		EndDialog( hWnd, 1 );
	DriverDisconnect();
	PostQuitMessage( 1 );
}
   

//----------------------------------------------------------------------
//  
// CallDriver
//
// Sends a DevIoControl to the driver along with some data if necessary.
//
//----------------------------------------------------------------------
BOOL CallDriver( int Msg, VOID * InData, int InDataLen, 
				VOID * OutData, int OutDataLen )
{
#if !TESTING
	int			nb;

	if ( ! DeviceIoControl(	sys_handle, Msg,
							InData, InDataLen, OutData, OutDataLen, &nb, NULL ) )
	{
		Abort( NULL, TEXT("Couldn't access device driver") );
		return FALSE;
	}
#endif
	return TRUE;
}


//----------------------------------------------------------------------
//  
// MS, NS
//
// Conversion routines between milliseconds and nanoseconds.
//
//----------------------------------------------------------------------
ULONG MS( LARGE_INTEGER t )
{
	/* t * 100ns -> ms */
	return (ULONG) (-t.QuadPart / 10000);
}

LARGE_INTEGER NS( ULONG t )
{
	/* t * ms -> 100ns */
	LARGE_INTEGER i;
	i.QuadPart = -10000i64 * t;
	return i;
}

 

//----------------------------------------------------------------------
//  
// BalloonDialog
//
// Dialog function for home-brewed balloon help
//
//----------------------------------------------------------------------
LONG APIENTRY BalloonDialog( HWND hDlg, UINT message, UINT wParam, LONG lParam )
{
	static struct tip *	ctx;
	static RECT			rect;
	static HFONT		hfont;
	LPCREATESTRUCT		lpcs;
	HDC					hdc;
	LOGFONT				lf;
	RECT				AssocRect;

	switch (message) {
		case WM_CREATE:
			lpcs = (void *)lParam;
			ctx = lpcs->lpCreateParams;
			hdc = GetDC( hDlg );

			//
			// Get font
			//
			if ( hfont == NULL )  {
				GetObject( GetStockObject(SYSTEM_FONT), sizeof lf, &lf ); 
				lf.lfWeight = FW_NORMAL;
				lf.lfHeight = 8;
				lf.lfWidth  = 0;
				strcpy( lf.lfFaceName, "MS Sans Serif" );
 				hfont = CreateFontIndirect( &lf ); 
			} 

			//
			// Compute size of required rectangle
			//
			rect.left	= 0;
			rect.top	= 0;
			rect.right	= lpcs->cx;
			rect.bottom	= lpcs->cy;
			SelectObject( hdc, hfont );
			DrawText( hdc, ctx->Text, -1, &rect, 
						DT_NOCLIP|DT_LEFT|DT_NOPREFIX|DT_WORDBREAK|DT_CALCRECT );

			//
			// Place below selected window
			//
			GetWindowRect( GetDlgItem( hMainDlg, ctx->ID ), &AssocRect );


			//
			// Move and resize window
			//
			MoveWindow( hDlg, AssocRect.right, AssocRect.top, 
						rect.right - rect.left + 10, 
						rect.bottom - rect.top + 10,
						TRUE );

			//
			// Adjust rectangle so text is centered
			//
			rect.left	+= 5;
			rect.right	+= 5;
			rect.top	+= 5;
			rect.bottom	+= 5;
			break;

		case WM_PAINT:
			hdc = GetDC( hDlg );

			//
			// Set colors
			//
			SetTextColor( hdc, 0x00000000 );
			SetBkMode( hdc, TRANSPARENT );
			SelectObject( hdc, hfont );
			DrawText( hdc, ctx->Text, -1, &rect, 
						DT_NOCLIP|DT_LEFT|DT_NOPREFIX|DT_WORDBREAK );
			break;

		case WM_MOUSEMOVE:
		case WM_CLOSE:	
			DestroyWindow( hDlg );
			break;
	}

    return DefWindowProc( hDlg, message, wParam, lParam );
}


//----------------------------------------------------------------------
//  
// DoGet
//
// Grab Cache Manager parameters from dialog controls.
//
//----------------------------------------------------------------------
BOOL DoGet( HWND hDlg, struct cacheman_getset * getset )
{
	if ( ! CallDriver( CACHEMAN_GET, NULL, 0, getset, sizeof *getset ) )
		return FALSE;

	SetDlgItemInt( hDlg, IDC_CCFIRSTDELAY,				MS(getset->CcFirstDelay),		FALSE );
	SetDlgItemInt( hDlg, IDC_CCIDLEDELAY,				MS(getset->CcIdleDelay),			FALSE );
	SetDlgItemInt( hDlg, IDC_CCCOLLISIONDELAY,			MS(getset->CcCollisionDelay),	FALSE );
	SetDlgItemInt( hDlg, IDC_CCTARGETCLEANDELAY,		MS(getset->CcTargetCleanDelay),	FALSE );
	SetDlgItemInt( hDlg, IDC_CCDIRTYPAGETHRESHOLD,		getset->CcDirtyPageThreshold,	FALSE );
	SetDlgItemInt( hDlg, IDC_CCDIRTYPAGETARGET,			getset->CcDirtyPageTarget,		FALSE );
	SetDlgItemInt( hDlg, IDC_CCAVAILABLEMAXDIRTYWRITE,	getset->CcMaxDirtyWrite,			FALSE );
	SetDlgItemInt( hDlg, IDC_CCAVAILABLEPAGESTHRESHOLD,	getset->CcAvailablePagesThreshold,FALSE );
	SetDlgItemInt( hDlg, IDC_MMSYSTEMCACHEWSMINIMUM,	getset->MmSystemCacheWsMinimum,	FALSE );
	SetDlgItemInt( hDlg, IDC_MMSYSTEMCACHEWSMAXIMUM,	getset->MmSystemCacheWsMaximum,	FALSE );

	CheckDlgButton( hDlg, IDC_MMDOPERIODICAGRESSIVETRIMMING,
					getset->MiDoPeriodicAgressiveTrimming != 0 );
	EnableWindow( GetDlgItem( hDlg, IDC_MMPERIODICAGRESSIVETRIMMINFREE ),
					getset->MiDoPeriodicAgressiveTrimming );
	EnableWindow( GetDlgItem( hDlg, IDC_MMPERIODICAGRESSIVETRIMMAXFREE ),
					getset->MiDoPeriodicAgressiveTrimming );
	EnableWindow( GetDlgItem( hDlg, IDC_MMPERIODICAGRESSIVECACHEWSMIN ),
					getset->MiDoPeriodicAgressiveTrimming );

	SetDlgItemInt( hDlg, IDC_MMPERIODICAGRESSIVETRIMMINFREE,
					getset->MmPeriodicAgressiveTrimMinFree, FALSE );
	SetDlgItemInt( hDlg, IDC_MMPERIODICAGRESSIVETRIMMAXFREE,
					getset->MmPeriodicAgressiveTrimMaxFree, FALSE );
	SetDlgItemInt( hDlg, IDC_MMPERIODICAGRESSIVECACHEWSMIN,
					getset->MmPeriodicAgressiveCacheWsMin, FALSE );

	SetDlgItemInt( hDlg, IDC_MMWORKINGSETREDUCTIONMAXCACHEWS,
					getset->MmWorkingSetReductionMaxCacheWs, FALSE );
	SetDlgItemInt( hDlg, IDC_MMWORKINGSETREDUCTIONMINCACHEWS,
					getset->MmWorkingSetReductionMinCacheWs, FALSE );
	SetDlgItemInt( hDlg, IDC_MMWORKINGSETVOLREDUCTIONMAXCACHEWS,
					getset->MmWorkingSetVolReductionMaxCacheWs, FALSE );

	EnableWindow( GetDlgItem( hDlg, IDC_SET ), TRUE );
	
	return TRUE;
}


//----------------------------------------------------------------------
//  
// DoSet
//
// Update dialog controls based on cache manager settings obtained
// from the device driver.
//
//----------------------------------------------------------------------
BOOL DoSet( HWND hDlg, struct cacheman_getset * getset )
{
	getset->CcFirstDelay			= NS(GetDlgItemInt( hDlg, IDC_CCFIRSTDELAY,				NULL,	FALSE ) );
	getset->CcIdleDelay				= NS(GetDlgItemInt( hDlg, IDC_CCIDLEDELAY,				NULL,	FALSE ) );
	getset->CcCollisionDelay		= NS(GetDlgItemInt( hDlg, IDC_CCCOLLISIONDELAY,			NULL,	FALSE ) );
	getset->CcTargetCleanDelay		= NS(GetDlgItemInt( hDlg, IDC_CCTARGETCLEANDELAY,		NULL,	FALSE ) );
	getset->CcDirtyPageThreshold	= GetDlgItemInt( hDlg, IDC_CCDIRTYPAGETHRESHOLD,		NULL,	FALSE );
	getset->CcDirtyPageTarget		= GetDlgItemInt( hDlg, IDC_CCDIRTYPAGETARGET,			NULL,	FALSE );
	getset->CcMaxDirtyWrite			= GetDlgItemInt( hDlg, IDC_CCAVAILABLEMAXDIRTYWRITE,	NULL,	FALSE );
	getset->CcAvailablePagesThreshold= GetDlgItemInt( hDlg, IDC_CCAVAILABLEPAGESTHRESHOLD,	NULL,	FALSE );
	getset->MmSystemCacheWsMinimum	= GetDlgItemInt( hDlg, IDC_MMSYSTEMCACHEWSMINIMUM,		NULL,	FALSE );
	getset->MmSystemCacheWsMaximum	= GetDlgItemInt( hDlg, IDC_MMSYSTEMCACHEWSMAXIMUM,		NULL,	FALSE );

	getset->MmPeriodicAgressiveTrimMinFree
		= GetDlgItemInt( hDlg, IDC_MMPERIODICAGRESSIVETRIMMINFREE, NULL, FALSE );

	getset->MmPeriodicAgressiveTrimMaxFree 
		= GetDlgItemInt( hDlg, IDC_MMPERIODICAGRESSIVETRIMMAXFREE, NULL, FALSE );

	getset->MmPeriodicAgressiveCacheWsMin
		= GetDlgItemInt( hDlg, IDC_MMPERIODICAGRESSIVECACHEWSMIN, NULL, FALSE );

	getset->MmWorkingSetReductionMaxCacheWs
		= GetDlgItemInt( hDlg, IDC_MMWORKINGSETREDUCTIONMAXCACHEWS, NULL, FALSE );

	getset->MmWorkingSetReductionMinCacheWs
		= GetDlgItemInt( hDlg, IDC_MMWORKINGSETREDUCTIONMINCACHEWS, NULL, FALSE );

	getset->MmWorkingSetVolReductionMaxCacheWs
		= GetDlgItemInt( hDlg, IDC_MMWORKINGSETVOLREDUCTIONMAXCACHEWS, NULL, FALSE );

	return CallDriver( CACHEMAN_SET, getset, sizeof *getset, NULL, 0 );
}


//----------------------------------------------------------------------
//  
// MainDialog
//
// This is the main window. 
//
//----------------------------------------------------------------------
LONG APIENTRY MainDialog( HWND hDlg, UINT message, UINT wParam, LONG lParam )
{
	static struct cacheman_getset		getset;
	HELPINFO						*	hi;
	int									j;
	
	switch (message) {
	case WM_INITDIALOG :
		if ( ! DriverConnect() )  {
			Abort( hDlg, "Unable to connect to device driver" );
			break;
		}

		//
		// An error will indicate that the build is not supported
		//
		if ( ! DoGet( hDlg, &getset ) )
			Abort( hDlg, "This build (version and/or service pack) of NT is not supported" );
		break;


	case WM_COMMAND :
		switch ( LOWORD( wParam ) ) {
		
		case IDOK :
		case IDCANCEL:
			PostMessage( hDlg, WM_CLOSE, 0, 0 );
			break;

		case IDC_MMDOPERIODICAGRESSIVETRIMMING:
 			getset.MiDoPeriodicAgressiveTrimming
				= IsDlgButtonChecked( hDlg, IDC_MMDOPERIODICAGRESSIVETRIMMING );
			EnableWindow( GetDlgItem( hDlg, IDC_MMPERIODICAGRESSIVETRIMMINFREE ),
							getset.MiDoPeriodicAgressiveTrimming );
			EnableWindow( GetDlgItem( hDlg, IDC_MMPERIODICAGRESSIVETRIMMAXFREE ),
							getset.MiDoPeriodicAgressiveTrimming );
			EnableWindow( GetDlgItem( hDlg, IDC_MMPERIODICAGRESSIVECACHEWSMIN ),
							getset.MiDoPeriodicAgressiveTrimming );
			break;


		case IDC_GET:
			if ( ! DoGet( hDlg, &getset ) )
				MessageBox( hDlg, "Error fetching parameters", NULL, MB_OK|MB_ICONEXCLAMATION );
			break;

		case IDC_SET:
			if ( ! DoSet( hDlg, &getset ) )
				MessageBox( hDlg, "Error fetching parameters", NULL, MB_OK|MB_ICONEXCLAMATION );
			break;

			break;

		}
		break;
		

	case WM_HELP:
		hi = (LPHELPINFO) lParam;
		for ( j = 0; j < sizeof Tips/sizeof Tips[0]; ++j )
			if ( hi->iCtrlId == Tips[j].ID )  {
				HWND h;
				h = CreateWindowEx( 0, "BALLOON", "balloon", 
									WS_POPUP|WS_VISIBLE|WS_BORDER,
									100, 100,
									200, 200,
									hDlg, NULL, hInst, 
									&Tips[j] );
			}
		break;


	case WM_CLOSE:	
		DriverDisconnect();
		EndDialog( hDlg, 0 );
		PostQuitMessage( 0 );
		return 0;
	}

    return DefWindowProc( hDlg, message, wParam, lParam);
}


//----------------------------------------------------------------------
//  
// WinMain
//
// Registers a class. The reason I did this is because the application
// doesn't get cleaned up properly upon exit if winmain just calls
// dialogbox. This was manifested by all the system icons disappearing
// after the program exited. See Petzold's hexcalc example for the base
// of this.
//
//----------------------------------------------------------------------
int CALLBACK WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
        				LPSTR lpCmdLine, int nCmdShow )
{	  
	MSG				msg;
	WNDCLASSEX		wndclass;


	//
	// save instance
	//
	hInst = hInstance;

	//
	// init common controls
	//
	InitCommonControls();

	//
	// create the main window class
	//
	wndclass.cbSize			= sizeof( WNDCLASSEX );
	wndclass.style          = CS_HREDRAW | CS_VREDRAW;
 	wndclass.lpfnWndProc    = (WNDPROC) MainDialog;
	wndclass.cbClsExtra     = 0;
	wndclass.cbWndExtra     = DLGWINDOWEXTRA;
	wndclass.hInstance      = hInstance;
	wndclass.hIcon          = LoadIcon( hInstance, "APPICON" );
	wndclass.hIconSm		= LoadIcon( hInstance, "APPICON" );
	wndclass.hCursor        = LoadCursor( NULL, IDC_ARROW );
	wndclass.hbrBackground  = (HBRUSH) (COLOR_BTNFACE+1);
	wndclass.lpszMenuName   = NULL;
	wndclass.lpszClassName  = "CACHEMAN";
	RegisterClassEx( &wndclass );

 	wndclass.lpfnWndProc    = (WNDPROC) BalloonDialog;
	wndclass.hbrBackground  = CreateSolidBrush( 0x0080FFFF );
	wndclass.lpszClassName  = "BALLOON";
	RegisterClassEx( &wndclass );

	//	
	// create the dialog
	//
	hMainDlg = CreateDialog( hInstance, "CACHEMAN", NULL, (DLGPROC)MainDialog );
	ShowWindow( hMainDlg, nCmdShow) ;
 
	while ( GetMessage( &msg, NULL, 0, 0 ) )  {
		if ( ! IsDialogMessage( hMainDlg, &msg ) )  {
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}
	}
	return msg.wParam;
}

